/* ------------------------------------------------------------------------- */
/* board/motorola_fads/flash.c                                               */
/*                                                                           */
/* Flash specific code for Motorola (F)ADS reference boards.                 */
/*                                                                           */
/* Copyright (C) 2000, 2001, 2002 Magnus Damm <damm@opensource.se>           */
/*                                                                           */
/* This program is free software; you can redistribute it and/or             */
/* modify it under the terms of the GNU General Public License as            */
/* published by the Free Software Foundation; either version 2 of            */
/* the License, or (at your option) any later version.                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software               */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston,                     */
/* MA 02111-1307 USA                                                         */
/* ------------------------------------------------------------------------- */

#include "types.h"
#include "board.h"
#include "cpu.h"
#include "console.h"
#include "8xx_immap.h"
#include "flash.h"
#include "string.h"

//#define DEBUG_FLASH

typedef struct {

  u32 simm_size;
  u32 first_sector;
  u8 no_banks;
  u8 no_sectors;
  u8 simm_manufacturer;

} board_flash_info_t;

#define FLASH_CODE_ADDRESS \
  ((void *) &((((immap_t *)mpc8xx_get_imap())->im_cpm.cp_dparam[ \
  ((sizeof(board_flash_info_t) + 4) & ~3) ])))

#define FLASH_INFO(imap) \
  ((board_flash_info_t *) &(((immap_t *)(imap))->im_cpm.cp_dparam[0]))

void board_flash_program_init(void);
void board_flash_setup(unsigned int bcsr2);

/* ------------------------------------------------------------------------- */
/* flash                                                                     */
/*                                                                           */
/* The motorola (F)ADS has the flash on a SIMM.                              */
/* All modules I've seen has been 4 x 8 bit devices -> 32 bits to the cpu.   */
/* According to the manual are there two kinds of chips used on the modules: */
/* AMD 29F040 and Intel 28F008. The manual does not say anything about how   */
/* the devices are organized, but my SIMM has AMD 29F040B chips on it.       */
/*                                                                           */

#define MINIBOOT_SIZE    16384 /* bytes */
#define FLASH_START      0

/* 29F040B is 512KByte per device */

const unsigned char sectors_29F040B[] = { 64,64,64,64,64,64,64,64 };

/* 28F008B is 1MByte per device */

const unsigned char sectors_28F008B[] = { 8, 8, 8, 8, 8, 8, 8, 8,
					  64,64,64,64,64,64,64,64,
					  64,64,64,64,64,64,64 };

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

void board_flash_setup(unsigned int bcsr2)
{
  u32 simm_size, simm_mf, ps;
  int ss, k;
  char *sp = NULL;
  board_flash_info_t *bfit = FLASH_INFO(mpc8xx_get_imap());

  /* read out FLASH_PD bits from BCSR2 */

#define MOTOROLA 1
#define SMART    2

  switch(bcsr2 >> 28) {
    
    /* 8MByte SIMMS */
    
  case 6: 
    simm_size = 8;
    simm_mf = MOTOROLA;
    break;
    
  case 4: 
    simm_size = 8;
    simm_mf = SMART;
    break;
    
    /* 4MByte SIMMS */
    
  case 7: 
    simm_size = 4;
    simm_mf = MOTOROLA;
    break;
    
  case 5: 
    simm_size = 4;
    simm_mf = SMART;
    break;
    
    /* 2MByte SIMM */
    
  case 8:
    simm_size = 2;
    simm_mf = MOTOROLA;
    break;
    
  default:
    simm_size = 8;            /* unknown - big for no limits */
    simm_mf = 0;
  }
  
  memset(bfit, 0, sizeof(board_flash_info_t));

  bfit->simm_size = simm_size << 20;  /* to Megabytes */
  bfit->simm_manufacturer = simm_mf;
    
  if (!simm_mf)
    return;
  
  /* figure out 
   * which sector is the first one 
   * that the user is allowed to write to.
   */
  
  ss = ps = 0;
  
  if (simm_mf == MOTOROLA) {
    ss = sizeof(sectors_29F040B);
    sp = (void *) sectors_29F040B;
    
    bfit->no_banks = bfit->simm_size / (2 << 20);  /* 4 * 512KByte */
    bfit->no_sectors = bfit->no_banks * sizeof(sectors_29F040B);
  }
  else
    if (simm_mf == SMART) {
      ss = sizeof(sectors_28F008B);
      sp = (void *) sectors_28F008B;
      
      bfit->no_banks = bfit->simm_size / (4 << 20);  /* 4 * 1MByte */
      bfit->no_sectors = bfit->no_banks * sizeof(sectors_28F008B);
    }
  
  for (k = 0; k < ss; k++) {
    ps += sp[k] << (10 + 2);  /* table in kbytes, 4 chips */
    
    if (ps >= MINIBOOT_SIZE)
      break;
  }
  
  bfit->first_sector = ps;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

void board_flash_getinfo(unsigned int *start, 
			 unsigned int *stop, 
			 unsigned int *number_of_sectors)
{
  board_flash_info_t *bfit = FLASH_INFO(mpc8xx_get_imap());

  *start = FLASH_START;
  *stop  = FLASH_START + bfit->simm_size;

  *number_of_sectors = bfit->no_sectors;

#ifdef DEBUG_FLASH
  printf("bfit dump:\n");
  printf("simm_size = 0x%08x\n", bfit->simm_size);
  printf("first_sector = 0x%08x\n", bfit->first_sector);
  printf("no_sectors = %d\n", bfit->no_sectors);
  printf("no_banks = %d\n", bfit->no_banks);
  printf("simm_manufacturer = %d\n", bfit->simm_manufacturer);
#endif

}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* flash_erase_sector will be copied to the internal DPRAM and execute there.*/
/* why? we cannot execute from flash while we are programming/erasing it.    */
/* so, don't try any to call anything at all, right? only inline.            */

#define WRITE_FLASH(a, d)       \
do                              \
{                               \
  *((volatile u32 *) (a)) = (d);\
 } while(0)

#define WRITE_FLASH8(a, d)      \
do                              \
{                               \
  *((volatile u8 *) (a)) = (d);\
 } while(0)

int board_flash_erase_sector_ram(immap_t *immap, 
				 volatile u32 *sector_address, 
				 volatile u32 *chip_address)
{
  board_flash_info_t *bfit = FLASH_INFO(immap);
  int c;
  u32 final_value;

  if (bfit->simm_manufacturer == MOTOROLA) {     

    /* Motorola SIMMS use AMD chips */

    /* first - autoselect and check if the sector is protected */

    WRITE_FLASH(&chip_address[0x5555], 0xAAAAAAAA);
    WRITE_FLASH(&chip_address[0x2aaa], 0x55555555);
    WRITE_FLASH(&chip_address[0x5555], 0x90909090);

    /* check if:
     * manufacturer code is ok
     * device id matches 29f040
     * protection status
     */

    c = FLASH_OK;

    if (chip_address[0] != 0x01010101)
      c = FLASH_BAD_MANUFACTURER_ID;
    else
      if (chip_address[1] != 0xa4a4a4a4)
	c = FLASH_BAD_DEVICE_ID;
      else
	if (sector_address[2] & 0x01010101)
	  c = FLASH_SECTOR_PROTECTED_ERROR;

    /* reset back to normal mode */

    WRITE_FLASH(&chip_address[0], 0xF0F0F0F0);
    
    if (c != FLASH_OK)
      return c;

    /* erase if everything is ok */

    WRITE_FLASH(&chip_address[0x5555], 0xAAAAAAAA);
    WRITE_FLASH(&chip_address[0x2aaa], 0x55555555);
    WRITE_FLASH(&chip_address[0x5555], 0x80808080);
    WRITE_FLASH(&chip_address[0x5555], 0xAAAAAAAA);
    WRITE_FLASH(&chip_address[0x2aaa], 0x55555555);

    WRITE_FLASH(&sector_address[0], 0x30303030);

    final_value = 0xffffffff;
  }
  else
    if (bfit->simm_manufacturer == SMART) {

      /* Smart SIMMS use Intel chips */

      /* clear status register */

      WRITE_FLASH(sector_address, 0x50505050); 

      /* block erase           */

      WRITE_FLASH(sector_address, 0x20202020); 
      WRITE_FLASH(sector_address, 0xd0d0d0d0); 

      final_value = 0x80808080;
      
    }
    else
      return FLASH_ERROR;

  for(c = 0; c < 2000; c++) {
    
    /* wait 1000 us and kick the watchdog */
    {
      volatile cpmtimer8xx_t *t =  &immap->im_cpmtimer;
      
      t->cpmt_tgcr &= 0xfff8; /* stop and reset timer (18-8) */
      
      t->cpmt_tmr1 = 0x001a;  
      t->cpmt_tcn1 = 0x0000;
      t->cpmt_trr1 = (MOT_FADS_HZ / 1000);  /* wrap warning!!! */
      t->cpmt_ter1 = 0xffff;
      
      t->cpmt_tgcr |= 0x0001; /* enable timer */
      
      do { /* poll event register while kicking the dog */
	
	immap->im_siu_conf.sc_swsr = 0x556c; /* (11-21) */
	immap->im_siu_conf.sc_swsr = 0xaa39;
	
      } while (!t->cpmt_ter1);
    }
    
    if (*sector_address == final_value) {
      c = 0;
      break;
    }
  }

  WRITE_FLASH(sector_address, 0xf0f0f0f0); 

  if (c)
    return FLASH_ERASE_ERROR;

  return 0;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

int board_flash_erase_sector(int s)
{
#define ERASE_SIZE ((u32)board_flash_erase_sector - (u32)board_flash_erase_sector_ram)

  u32 dummy[ERASE_SIZE/4];
  board_flash_info_t *bfit = FLASH_INFO(mpc8xx_get_imap());
  int (*f)(immap_t *, u32 *, u32 *) = (void *) dummy;
  u32 sector_address, chip_address;
  int ret;
  int ss, k;
  char *sp = NULL;

  memcpy (dummy, board_flash_erase_sector_ram, ERASE_SIZE);

  if (bfit->simm_manufacturer == MOTOROLA) {
    ss = sizeof(sectors_29F040B);
    sp = (void *) sectors_29F040B;
  }
  else
    if (bfit->simm_manufacturer == SMART) {
      ss = sizeof(sectors_28F008B);
      sp = (void *) sectors_28F008B;
    }
    else
      return FLASH_ERROR;

#ifdef DEBUG_FLASH
  printf("erase: s = %d, ss = %d\n", s, ss);
#endif  

  sector_address = chip_address = FLASH_START;

  for (k = 0; k < s; k++) {
    if (k == ss) {
      chip_address = sector_address;
      s -= ss;
      k = 0;
    }

    sector_address += sp[k] << (10 + 2); /* table in kbytes, 4 chips */
  }

#ifdef DEBUG_FLASH
  printf("erase: sector_address = 0x%08x\n", sector_address);
  printf("erase: chip_address = 0x%08x\n", chip_address);
#endif  

  if (sector_address < bfit->first_sector) /* do not erase reserved sectors */
    return FLASH_OK;

  if (sector_address >= bfit->simm_size)
    return FLASH_BAD_ADDRESS;

  mpc8xx_icache_invalidate();

  ret = f(mpc8xx_get_imap(), 
	  (u32 *) sector_address, 
	  (u32 *) chip_address);
  
  mpc8xx_icache_invalidate();

  return ret;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* this code will also be copied to the internal DPRAM and executed there.   */

int board_flash_program_ram(immap_t *immap, 
			    volatile u32 *address, 
			    u32 data, 
			    volatile u32 *chip_address)
{
  board_flash_info_t *bfit = FLASH_INFO(immap);
  int c;
  u32 final_value;

  if (bfit->simm_manufacturer == MOTOROLA) {

    /* Motorola SIMMS use AMD chips */

    WRITE_FLASH(&chip_address[0x5555], 0xAAAAAAAA);
    WRITE_FLASH(&chip_address[0x2aaa], 0x55555555);
    WRITE_FLASH(&chip_address[0x5555], 0xA0A0A0A0);

    WRITE_FLASH(address, data);

    final_value = data;
  }
  else
    if (bfit->simm_manufacturer == SMART) {

      /* Smart SIMMS use Intel chips */

      /* program */
      
      WRITE_FLASH(address, 0x40404040); 
      WRITE_FLASH(address, data); 

      final_value = 0x80808080;
    }
    else
      return FLASH_ERROR;

  /* timeout: one second */

  for(c = 0; c < 100000; c++) {
    /* wait 10 us and kick the watchdog */

    volatile cpmtimer8xx_t *t =  &immap->im_cpmtimer;

    t->cpmt_tgcr &= 0xfff8; /* stop and reset timer (18-8) */

    t->cpmt_tmr1 = 0x001a;  
    t->cpmt_tcn1 = 0x0000;
    t->cpmt_trr1 = (MOT_FADS_HZ / 100000);  /* wrap warning!!! */
    t->cpmt_ter1 = 0xffff;

    t->cpmt_tgcr |= 0x0001; /* enable timer */

    do { /* poll event register while kicking the dog */

      immap->im_siu_conf.sc_swsr = 0x556c; /* (11-21) */
      immap->im_siu_conf.sc_swsr = 0xaa39;
      
    } while (!t->cpmt_ter1);

    if (*address == final_value) {
      c = 0;
      break;
    }
  }

  WRITE_FLASH(address, 0xf0f0f0f0); 

  if (c)
      return FLASH_PROGRAM_ERROR;

  return FLASH_OK;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

void board_flash_program_init(void)
{
#define PROGRAM_SIZE ((u32)board_flash_program - (u32)board_flash_program_ram)

  memcpy ((void *) FLASH_CODE_ADDRESS, board_flash_program_ram, PROGRAM_SIZE);

  mpc8xx_icache_invalidate();
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

static int board_flash_check_access(u32 *address)
{
  board_flash_info_t *bfit = FLASH_INFO(mpc8xx_get_imap());
  u32 a = (u32) address;

#ifdef DEBUG_FLASH
  printf("check: attempt to access 0x%08x\n", a);
#endif  

  if (a > (bfit->simm_size - 3))
    return FLASH_BAD_ADDRESS;

  if (a < bfit->first_sector)
    return FLASH_BAD_ADDRESS;
  
#ifdef DEBUG_FLASH
  printf("check: address ok: 0x%08x\n", (u32) a);
#endif  

  return FLASH_OK;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

int board_flash_program(u32 *address, u32 data)
{
  int (*f)(immap_t *, volatile u32 *, u32, u32 *) =  FLASH_CODE_ADDRESS;
  board_flash_info_t *bfit = FLASH_INFO(mpc8xx_get_imap());
  u32 chip_address, bankmask;
  int ret;

  ret = board_flash_check_access(address);

  if (ret == FLASH_OK) {

    bankmask = (bfit->simm_size / bfit->no_banks) - 1;
    chip_address = (((u32) (address)) - FLASH_START) & ~bankmask;
    chip_address += FLASH_START;

    ret = f(mpc8xx_get_imap(), 
	    (volatile u32 *) address, 
	    data, 
	    (volatile u32 *) chip_address);
    
    mpc8xx_icache_invalidate();
    
#ifdef DEBUG_FLASH
    printf("program: address = 0x%08x read out = 0x%08x, data = 0x%08x\n", 
	 (u32) address, *address, data);
#endif  
  }

  return ret;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

int board_flash_read(u32 *address, u32 *data)
{
  int ret;

  ret = board_flash_check_access(address);

  if (ret == FLASH_OK)
    *data = *address;

  return ret;
}

/* ------------------------------------------------------------------------- */
