/* ------------------------------------------------------------------------- */
/* board/switchcore_rd2010/flash.c                                           */
/*                                                                           */
/* Flash code for the "rd2010" reference board designed by Switchcore.       */
/*                                                                           */
/* Copyright (C) 2000, 2001, 2002 Magnus Damm <damm@opensource.se>           */
/*                                                                           */
/* This program is free software; you can redistribute it and/or             */
/* modify it under the terms of the GNU General Public License as            */
/* published by the Free Software Foundation; either version 2 of            */
/* the License, or (at your option) any later version.                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software               */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston,                     */
/* MA 02111-1307 USA                                                         */
/* ------------------------------------------------------------------------- */

#include "types.h"
#include "board.h"
#include "cpu.h"
#include "console.h"
#include "string.h"
#include "flash.h"

#include "8xx_immap.h"

#define FLASH_CODE_ADDRESS \
  ((u32) &((((immap_t *)mpc8xx_get_imap())->im_cpm.cp_dparam[0])))

/* parameters for the 28F640J3A chip:
 *
 * bus size:    16 bit
 * sector size: 128 Kbytes
 * chip size:   8 Mbytes
 */

#define SECTORSIZE       (128 << 10)
#define RESERVED_SECTORS_BOTTOM 1 /* 1 sector for miniboot           */
#define RESERVED_SECTORS_TOP    5 /* 5 sectors for user/factory data */
#define FLASH_START      (0 + (RESERVED_SECTORS_BOTTOM * SECTORSIZE))
#define FLASH_END        ((8 << 20) - (RESERVED_SECTORS_TOP * SECTORSIZE))
#define FLASH_NOSECTORS  ((FLASH_END - FLASH_START) / SECTORSIZE)

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

void board_flash_getinfo(unsigned int *start, 
			 unsigned int *stop,
			 unsigned int *number_of_sectors)
{
  *start = FLASH_START;
  *stop  = FLASH_END;
  
  *number_of_sectors = FLASH_NOSECTORS;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

void board_flash_getsectorinfo(unsigned int sector, 
			       unsigned char**base, unsigned int *size)
{
  if (sector < FLASH_NOSECTORS) {
    *base = (void *)(FLASH_START + (sector * SECTORSIZE));
    *size = SECTORSIZE;
  }
  else {
    *base = 0;
    *size = 0;
  }
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* functions below will be copied to the internal DPRAM and execute there.   */
/* why? we cannot execute from flash while we are programming/erasing it.    */
/* so, don't try any to call anything at all, right? only inline.            */

#define WRITE_FLASH(a, d)       \
do                              \
{                               \
  *((volatile u16 *) (a)) = (d);\
 } while(0)

int board_flash_erase_sector_ram(immap_t *immap, volatile u16 *address)
{
  int c;
 
  /* clear status register */

  WRITE_FLASH(address, 0x0050); 

  /* block erase           */

  WRITE_FLASH(address, 0x0020); 
  WRITE_FLASH(address, 0x00d0); 

  for(c = 0; c < 2000; c++) {

    /* wait 1000 us and kick the watchdog */
    {
      volatile cpmtimer8xx_t *t =  &immap->im_cpmtimer;

      t->cpmt_tgcr &= 0xfff8; /* stop and reset timer (18-8) */

      t->cpmt_tmr1 = 0x001a;  
      t->cpmt_tcn1 = 0x0000;
      t->cpmt_trr1 = (CPU_HZ / 1000);  /* wrap warning!!! */
      t->cpmt_ter1 = 0xffff;

      t->cpmt_tgcr |= 0x0001; /* enable timer */

      do { /* poll event register while kicking the dog */

	immap->im_siu_conf.sc_swsr = 0x556c; /* (11-21) */
	immap->im_siu_conf.sc_swsr = 0xaa39;
	
      } while (!t->cpmt_ter1);
    }

    if (*address == 0x0080) {
      c = 0;
      break;
    }
  }

  /* read array */

  WRITE_FLASH(address, 0x00ff); 

  if (c) {
    return FLASH_ERASE_ERROR;
  }

  return FLASH_OK;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/* this code will also be copied to the internal DPRAM and executed there.   */

int board_flash_program_ram(immap_t *immap, volatile u16 *address, u16 data)
{
  int c;

  /* program */

  WRITE_FLASH(address, 0x0040); 
  WRITE_FLASH(address, data); 

  /* timeout: one second */

  for(c = 0; c < 100000; c++) {
    /* wait 10 us and kick the watchdog */

    volatile cpmtimer8xx_t *t =  &immap->im_cpmtimer;

    t->cpmt_tgcr &= 0xfff8; /* stop and reset timer (18-8) */

    t->cpmt_tmr1 = 0x001a;  
    t->cpmt_tcn1 = 0x0000;
    t->cpmt_trr1 = (CPU_HZ / 100000);  /* wrap warning!!! */
    t->cpmt_ter1 = 0xffff;

    t->cpmt_tgcr |= 0x0001; /* enable timer */

    do { /* poll event register while kicking the dog */

      immap->im_siu_conf.sc_swsr = 0x556c; /* (11-21) */
      immap->im_siu_conf.sc_swsr = 0xaa39;
      
    } while (!t->cpmt_ter1);
    
    if (*address == 0x0080) {
      c = 0;
      break;
    }
  }

  /* read array */

  WRITE_FLASH(address, 0x00ff); 

  if (c) {
      return FLASH_PROGRAM_ERROR;
  }

  return FLASH_OK;
}


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

void board_flash_init(void)
{
#define ERASE_SIZE \
 ((u32) board_flash_program_ram - (u32) board_flash_erase_sector_ram)
#define PROGRAM_SIZE \
 ((u32) board_flash_init - (u32) board_flash_program_ram)
#define ERASE_ADDR FLASH_CODE_ADDRESS
#define PROGRAM_ADDR (FLASH_CODE_ADDRESS + ERASE_SIZE)

  memcpy ((void *) ERASE_ADDR, board_flash_erase_sector_ram, ERASE_SIZE);
  memcpy ((void *) PROGRAM_ADDR, board_flash_program_ram, PROGRAM_SIZE);

  mpc8xx_icache_invalidate();
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

int board_flash_erase_sector(int s)
{
  int (*f)(immap_t *, volatile u32 *) = (void *) ERASE_ADDR;
  volatile u32 *address;
  int stat = FLASH_OK;

  /* do not erase the reserved sectors */

  if (s < FLASH_NOSECTORS) {

    address = (volatile u32 *)(FLASH_START + (s * SECTORSIZE));

    /* if the address is outside the flash, we're finished. */

    if (((u32) address) >= FLASH_END) {
      return FLASH_BAD_ADDRESS;
    }

#ifdef DEBUG_FLASH
    printf("erasing at address 0x%08x\n", (u32) address);
#endif

    stat = f(mpc8xx_get_imap(), address);

#ifdef DEBUG_FLASH
    printf("done erasing at address 0x%08x\n", (u32) address);
#endif
  }

  return stat;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

int board_flash_program(u16 *address, u16 data)
{
  int (*f)(immap_t *, volatile u16 *, u16) = (void *) PROGRAM_ADDR;
  int ret;

  if ((((u32) address) & 0x01) || (((u32) address) >= FLASH_END)
      || (((u32) address) < FLASH_START)) {
    return FLASH_BAD_ADDRESS;
  }

  ret =  f(mpc8xx_get_imap(), (volatile u16 *) address, data);

  return ret;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

int board_flash_read(u16 *address, u16 *data)
{
  if ((((u32) address) & 0x01) || (((u32) address) >= FLASH_END)
      || (((u32) address) < FLASH_START)) {
    return FLASH_BAD_ADDRESS;
  }

  *data = *address;

  return FLASH_OK;
}

/* ------------------------------------------------------------------------- */
