/* ------------------------------------------------------------------------- */
/* lib/image.c                                                               */
/*                                                                           */
/* code handling images in flash.                                            */
/*                                                                           */
/* Copyright (C) 2000, 2001 Magnus Damm <damm@opensource.se>                 */
/* Copyright (C) 1999 - 2000 Ericsson Radio Systems AB                       */
/*                                                                           */
/* This program is free software; you can redistribute it and/or             */
/* modify it under the terms of the GNU General Public License as            */
/* published by the Free Software Foundation; either version 2 of            */
/* the License, or (at your option) any later version.                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software               */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston,                     */
/* MA 02111-1307 USA                                                         */
/* ------------------------------------------------------------------------- */

#include "types.h"
#include "cpu.h"
#include "board.h"
#include "console.h"
#include "string.h"

#include "image.h"
#include "crc16.h"

const u32 image_const = 0xffffffff;

int image_find(void **image)
{
  unsigned int start, stop, dummy;
  u32 *fl;
  image_header_t *ih;
  u32 nobytes;
  u16 crc;
  int k, m, s, r, inc, do_kick = 0;
  
  board_flash_getinfo(&start, &stop, &dummy);

  nobytes = stop - start - sizeof(image_header_t);

  fl = (unsigned int *) start;

  /* Speed up search by jumping to locations where
     images can be located */
#ifdef IMAGE_OFFSET
  inc = IMAGE_OFFSET/4;
#else
  inc = 1;
#endif
  
  for(k = 0; k < (nobytes >> 2);) {
    
    if (do_kick++ >= 0x10000) {
      cpu_watchdog_kick();
      do_kick = 0;
    }
    
    if (fl[k] == IMAGE_COOKIE) {
      ih = (image_header_t *) &fl[k];

      if (((nobytes-(k << 2)) >= ih->header_size + ih->data_size)) {

	/* calculate crc from start to (but not including) the crc16 field */

	crc = crc16_calc((u8 *)ih, (((u32) &ih->crc16_checksum) - ((u32) ih)), 
			 (u16) image_const);

	/* add 0xffffffff to crc */

	crc = crc16_calc((u8 *) &image_const, sizeof(u32), crc);

	cpu_watchdog_kick();

	/* calculate crc on the rest of the image */

	m = 0;
	s = (ih->header_size + ih->data_size) - 
	  ((((u32) &ih->crc16_checksum) - ((u32) ih)) + sizeof(u32));
	 
#define BLOCK_SIZE 0x10000
 
	while (m < s) {

	  if ((s - m) < BLOCK_SIZE)
	    r = s - m;
	  else
	    r = BLOCK_SIZE;

	  crc = crc16_calc((u8 *) ((u32)&ih->crc16_checksum) + sizeof(u32) + m,
			   r, crc);

	  cpu_watchdog_kick();

	  m += r;

	}

	if(crc == (u16) ih->crc16_checksum) {
	  *image = (void *) ih;
	  return 0;
	}

	cpu_watchdog_kick();
      }
      
    }
    
    k += inc;
  }

  return -1;
}

#define BUFLEN 65

void image_display_info(void *image)
{
  image_header_t *ih = image;
  u8 buf[BUFLEN];

  if (ih->name[0]) {
    memcpy(buf, ih->name, BUFLEN);
    buf[BUFLEN-1] = 0;
    printf("name:        \"%s\"\n", buf);
  }

  if (ih->description[0]) {
    memcpy(buf, ih->description, BUFLEN);
    buf[BUFLEN-1] = 0;
    printf("description: \"%s\"\n", buf);
  }

  printf("address:     0x%08x\n", (u32) image);
  printf("entry:       0x%08x\n", ((u32) image) + ih->header_size+ENTRY_OFFSET);
}

void image_get_startaddress(void *image, u32 *startaddress)
{
  image_header_t *ih = image;

  *startaddress = (((u32) image) + ih->header_size+ENTRY_OFFSET);
}

void image_get_name(void *image, u8 *buf, int bufsize)
{
  image_header_t *ih = image;

  memcpy(buf, ih->name, bufsize);
  buf[bufsize-1] = 0;
}






