/* ------------------------------------------------------------------------- */
/* main.c                                                                    */
/*                                                                           */
/* high level code for the Miniboot project.                                 */
/*                                                                           */
/* Copyright (C) 2000, 2001, 2002 Magnus Damm <damm@opensource.se>           */
/*                                                                           */
/* This program is free software; you can redistribute it and/or             */
/* modify it under the terms of the GNU General Public License as            */
/* published by the Free Software Foundation; either version 2 of            */
/* the License, or (at your option) any later version.                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software               */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston,                     */
/* MA 02111-1307 USA                                                         */
/* ------------------------------------------------------------------------- */

#include "types.h"
#include "cpu.h"
#include "board.h"
#include "console.h"
#include "image.h"
#include "srec.h"
#include "flash.h"

const char boot_message[] = "Miniboot v0.03";

/* ------------------------------------------------------------------------- */

static int wait_for_space(void)
{
  int i;

  for (i = 0; i < BOOT_DELAY; i++) {
    if (tstc() && (getc() == ' '))
      return 1;

    cpu_udelay(1000);
  }

  return 0;
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#define MAX_LINE 16
#define MAX_ARGS 2

void command_mode(void)
{
  char linebuf[MAX_LINE+1];
  char *arg[MAX_ARGS];
  int arg_nr;
  int cnt, i, j, display_prompt;
  char c;

  display_prompt = 1;

  while(1) {

    /* let the input buffer drain first */

    while(tstc())
      getc();
    
    /* display prompt */
    
    if (display_prompt)
      printf("> ");
    else
      display_prompt++;

    cnt = 0;

    /* read line */ 

    while(1) {
      c = getc();

      /* handle return */
      
      if (c == '\r') {
	putc(c);
	putc('\n');
	break;
      }
      
      /* handle backspace */
      
      if (c == 8) {
	if (cnt > 0) {
	  putc(c);
	  putc(' ');
	  putc(c);
	  cnt--;
	}
      }
      else
	
	/* insert the rest of the chars in our buffer */
	
	if (cnt < MAX_LINE) {
	  putc(c);
	  linebuf[cnt++] = c;
	}
    }

    /* zero out the unused part */

    while(cnt <= MAX_LINE)
      linebuf[cnt++] = 0;

    /* figure out arguments */
      
    for (i = 0; i < MAX_ARGS; i++)
      arg[i] = NULL;
    
    j = 0;
    arg_nr = 0;
    
    for (i = 0; i < MAX_LINE; i++) {
      if (linebuf[i] == ' ')
	linebuf[i] = 0;

      switch(j) {
      case 0: 
	if (linebuf[i] == 0)
	  break;
	else
	  j++;
	
	/* fall through */
	
      case 1:
	if (arg_nr < MAX_ARGS) {
	  arg[arg_nr++] = &linebuf[i];
	  j++;
	}
	else
	  j = -1;
	
	break;
	
      case 2:
	if (linebuf[i] == 0)
	  j = 0;
	
	break;
	
      case -1:
	break;
      }
    }	  
	
    /* check if arguments means something */

    if (arg[0]) {

      switch(*arg[0]) {
      case 'H':
      case 'h':
      case '?':
	printf("Help:\n");
	printf("B bps      - Change bps.\n");
	printf("C          - Check image.\n");
	printf("U          - Update image via serial port.\n");
	printf("R          - Reboot.\n");
#ifdef SUPPORT_CMD_Z
	printf("Z address  - Write zero at address (decimal).\n");
#endif
	break;

#ifndef BPS_MAX
#define BPS_MAX 115200
#endif

      case 'B':
      case 'b':
	{
	  int ok = 0;

	  if (arg[1]) {
	    i = 0;
	    j = 0;

	    ok = 1;

	    while (ok) {
	      
	      c = arg[1][i];

	      if ((c >= '0') && (c <= '9')) {
		j = j * 10;
		j = j + (c - '0');
		i++;
	      }
	      else {
		if (c == 0)
		  break;
		else
		  ok = 0;
	      }
	    }
	  }
	   
	  if ((!ok) || (j > BPS_MAX) || (j < 9600)) {
 	    printf("Usage: b bps\n");
 	    printf("supported bps: 9600 -> %d.\n", BPS_MAX);
	    break; 
	  }

	  printf("Bps will now be changed to %u.\n", j);

	  console_shutdown();

	  console_init(j);

	  display_prompt = 0;

	}
      break;
      
      case 'U':
      case 'u':
	{
	  u32 k, number_of_sectors;


	  printf("I need to erase the flash before upload is possible.\n");

	  board_flash_getinfo(&k, &k, &number_of_sectors);

	  i = FLASH_ERROR;

	  for (k = 0; k < number_of_sectors; k++) {
	    printf("Erasing, %d%c...", (100*k)/number_of_sectors, '%');
	    i = flash_erase_sector(k);
	    printf("\r");
	    if (i != FLASH_OK)
	      break;
	  }

	  if (flash_verbose(i) == 0) {
	    printf("Erasing, 100%% done.\n");
	    printf("Ready for upload - waiting for image.\n");
	    
	    srec_parse();
	  }
	}
	break;
      case 'C':
      case 'c':
	{
	  void *image;

	  if (image_find(&image) < 0)
	    printf("No image found.\n");
	  else {
	    image_display_info(image);
	  }
	}
	break;

#ifdef SUPPORT_CMD_Z
      case 'Z':
      case 'z':
	{
	  int ok = 0;

	  if (arg[1]) {
	    i = 0;
	    j = 0;

	    ok = 1;

	    while (ok) {
	      
	      c = arg[1][i];

	      if ((c >= '0') && (c <= '9')) {
		j = j * 10;
		j = j + (c - '0');
		i++;
	      }
	      else {
		if (c == 0)
		  break;
		else
		  ok = 0;
	      }
	    }
	  }
	   
	  if (!ok) {
 	    printf("Usage: Z address\n");
	    break; 
	  }

	  printf("Writing zero at address %d\n",j);
	  
	  if (flash_verbose(board_flash_program((void*)j,0)) == 0) {
	    printf("Erase done\n");
	  }
	}
	break;
#endif /* SUPPORT_CMD_Z */

      case 0:
	break;

      case 'R':
      case 'r':
	cpu_start();
	/* not reached */
	
      default:
	printf("Unknown command.\n");
      }
    }
  }
}

/* ------------------------------------------------------------------------- */
/* This is the first C-code running on our machine!                          */

int main(void)
{
  char *board;
  char buf[64];
  void *image = NULL;
  int do_command_mode = 0;
  int tmp;
  u32 startaddr;

  board_init(&board);                /* start the board-specific code ASAP */
                                     /* to let it kick the watchdog etc.   */

  console_init(0);                   /* it's always nice with some output  */

  printf("\n\n%s%s, " __DATE__ " " __TIME__ ".\n", boot_message, board);

  do_command_mode = wait_for_space();

  if (!do_command_mode) {
    tmp = image_find(&image);

    /* it counts if someone pressed space while we searched for images */

    while(tstc())
      if (getc() == ' ')
	do_command_mode = 1;
    
    if (!do_command_mode && (tmp < 0)) {
      printf("No valid image found.\n");
      do_command_mode = 1;
    }
  }

  if (do_command_mode)
    command_mode();

  image_get_name(image, buf, sizeof(buf));
  image_get_startaddress(image, &startaddr);

  printf("Booting \"%s\" at 0x%08x.\n", buf, startaddr);

  console_shutdown();

  board_cleanup();
  
  cpu_image_start(startaddr);

  return 0;  /* never reached */
}

/* ------------------------------------------------------------------------- */

