/* ------------------------------------------------------------------------- */
/* based on lib/image.c                                                      */
/*                                                                           */
/* Checks that a file contains a valid miniboot image                        */
/* Should be endian safe.                                                    */
/*                                                                           */
/* Copyright (C) 2001 Thomas Lange, <thomas@corelatus.com>                   */
/* Copyright (C) 2000, 2001 Magnus Damm <damm@opensource.se>                 */
/* Copyright (C) 1999 - 2000 Ericsson Radio Systems AB                       */
/*                                                                           */
/* This program is free software; you can redistribute it and/or             */
/* modify it under the terms of the GNU General Public License as            */
/* published by the Free Software Foundation; either version 2 of            */
/* the License, or (at your option) any later version.                       */
/*                                                                           */
/* This program is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             */
/* GNU General Public License for more details.                              */
/*                                                                           */
/* You should have received a copy of the GNU General Public License         */
/* along with this program; if not, write to the Free Software               */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston,                     */
/* MA 02111-1307 USA                                                         */
/* ------------------------------------------------------------------------- */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <malloc.h>

#include <asm/byteorder.h>

#include <sys/types.h>
#include <sys/stat.h>

typedef unsigned int u32;
typedef unsigned short u16;
typedef unsigned char u8;

#include "../lib/image.h"

const u32 image_const = 0xffffffff;

/* all values stored in big endian format */

int static image_check(void *image)
{
  u32 *fl;
  image_header_t *ih;
  u16 crc;
  int m, s, r;
  
  fl = (unsigned int *) image;
  
  if (fl[0] == IMAGE_COOKIE) {
    ih = (image_header_t *) &fl[0];

    /* calculate crc from start to (but not including) the crc16 field */
    
    crc = crc16_calc((u8 *)ih, (((u32) &ih->crc16_checksum) - ((u32) ih)), 
		     (u16) image_const);
    
    /* add 0xffffffff to crc */
    
    crc = crc16_calc((u8 *) &image_const, sizeof(u32), crc);
    
    /* calculate crc on the rest of the image */
    
    m = 0;
    s = (__be32_to_cpu(ih->header_size) + __be32_to_cpu(ih->data_size)) - 
      ((((u32) &ih->crc16_checksum) - ((u32) ih)) + sizeof(u32));
    
#define BLOCK_SIZE 0x10000
    
    while (m < s) {
      
      if ((s - m) < BLOCK_SIZE)
	r = s - m;
      else
	r = BLOCK_SIZE;
      
      crc = crc16_calc((u8 *) ((u32)&ih->crc16_checksum) + sizeof(u32) + m,
		       r, crc);
      
      m += r;
      
    }
    
    if(crc == (u16) __be32_to_cpu(ih->crc16_checksum)) {
      return 0;
    }
  } /* if */
return(-1);
}

static int image_file_check( char *file, int offset, int max_size ){
  int fd;
  char *buff;
  int res=-1;
  u32 bytes=0;

  buff=malloc(max_size);
  
  if((fd=open(file,O_RDONLY))<0){
    printf("Open of %s failed:",file);
    perror("");
    free(buff);
    return(-1);
  }

  if(lseek(fd,offset,SEEK_SET) < 0){
    perror("seek failed");
  } 
  else{
    if((bytes=read(fd,buff,max_size))<0){
      perror("read failed\n");
    }
    else{
      if(image_check(buff))
	res = -1;
      else
	res = 0;
    }
  }
  
  close(fd);
  free(buff);
  
  return(res);
}

int main(int argc, char *argv[]){
  /* Offset and maxsize are needed to access a generic
     device such as a flash device driver */
  u32 offset;
  u32 max;
  
  if(argc!=4){
    printf("Usage: check_file FILE OFFSET MAXSIZE\n");
    exit(-1);
  }
  
  offset = strtoul(argv[2],NULL,10);
  
  max = strtoul(argv[3],NULL,10);
  
  if(image_file_check( argv[1], offset, max )){
    printf("Error\n");
    return(-1);
  }
  printf("OK\n");
  return(0);
}

